// Program for calculating a transitive closure using the Warshall's
// Algorithm in order to determine which procedures being compiled may
// have side effects (or more importantly, which ones definitely don't,
// since external procedures not in this file (or the perms) are of
// unknown status with respect to side effects.)  Using the transitive
// closure of "A is called by B" for all calls finds all the procedures
// that could potentially call "A" and experience it's side-effects.

/*
 To detect whether a function call causes a side effect, the body of the
 function must be analyzed. A function call can produce a side effect in
 a number of ways, such as: modifying a global variable, modifying an
 internal static variable, modifying one of its pass-by-reference
 arguments, producing output, consuming input etc.

We consider the following causes of side effects through a function call:
 1) Performing I/O
 2) Modifying global variables
 3) Modifying local permanent variables (like static variables in C)
 4) Modifying an argument passed by reference
 5) Modifying a local variable, either automatic or static,
    of a function higher up in the function call sequence (usually via a pointer).

 -- http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.70.2096&rep=rep1&type=pdf
 */
#include <stdio.h>

#define MAX_PROCS 128
static int procno[MAX_PROCS]; // Maximum possible number of procedures
                              // in a compilation unit (file).  We *could*
                              // make this a re-sizable dynamic array off
                              // the heap, but that's overkill.

int next_free_procno = 0;     // Number of vertices in the graph

unsigned char callgraph[MAX_PROCS][MAX_PROCS];
unsigned char side_effect[MAX_PROCS];

int map(int sparse) {
  int mapped;
  procno[next_free_procno] = sparse;
  for (mapped = 0; mapped <= next_free_procno; mapped++) {
    if (procno[mapped] == sparse) {
      if (mapped == next_free_procno) next_free_procno++;
      return mapped;
    }
  }
}

void call(int caller, int callee) {
  int mcr, mce;
  mcr = map(caller);
  mce = map(callee);
  callgraph[mce][mcr] = 1;
  callgraph[mcr][mcr] = 1; // all procedures are considered to call themselves
  callgraph[mce][mce] = 1;
  printf("proc %0d calls proc %0d\n", caller, callee);
}

void has_side_effects(int proc) {
  side_effect[map(proc)] = 1;
  printf("proc %0d is known to have side effects\n", proc);
}

void show_procs_with_side_effects(void) {
  int i, j;
  printf("\nThe following matrix is the transitive closure of the call graph\n\n          ");
  for (i = 0; i < next_free_procno; i++) printf ("%0d ", procno[i]);
  printf ("\n");
  for (i = 0; i < next_free_procno; i++) {
    if (side_effect[i]) printf("BAD:  %0d: ",procno[i]); else printf("GOOD: %0d: ",procno[i]);
    for (j = 0; j < next_free_procno; j++) {
      if (side_effect[i]) side_effect[j] |= callgraph[i][j];
      printf ("%2d ", callgraph[i][j]);
    }
    printf("\n");
  }
  printf("\nThe following procedures could all have side effects, whether directly or indirectly:\n\n");
  for (int i = 0; i < next_free_procno; i++)
    if (side_effect[i]) printf("%d\n", procno[i]);
}

void Warshall(void) {
  int i, j, k;
  for (k = 0; k < next_free_procno; k++)
    for (i = 0; i < next_free_procno; i++)
      for (j = 0; j < next_free_procno; j++)
        callgraph[i][j] |= (callgraph[i][k] & callgraph[k][j]);
}

int main(int argc, char **argv) {
  // The internal ID numbers used for procedures in the compiler are
  // just the indexes of the tuples for the procedure definition,
  // which will be quite sparse in practice.

  call(10,13); // Call this every time a call to another procedure
  call(10,11); // is detected while compiling a procedure.
  call(11,12);
  call(12,13);

  has_side_effects(12); // Call this every time (or at least once) while
                        // compiling a procedure if a direct side-effect
                        // is detected. Note that we exclude procedure calls
                        // from this requirement.

  Warshall();                      // Discover every procedure that calls procedures
                                   // with side-effects, transitively.  Hence why we
                                   // need a transitive closure.

  show_procs_with_side_effects();  // Print the solution

  return 0;
}
