/*
    See also:
              https://www.csie.ntu.edu.tw/~b93501005/slide5.pdf
              https://www.mkssoftware.com/docs/man5/symbol.5.asp
              http://www-acaps.cs.mcgill.ca/~hendren/520/2016/slides/symbol.pdf
              https://opensource.apple.com/source/swig/swig-4/swig/Source/Swig/symbol.c.auto.html

 */
#include <stdio.h>
extern FILE *stderr; // when done testing rename stderr to stderr
int current_scope = -1;  // Base scope is scope 0. Most recent scope (top level) is current_scope.

#ifndef _SYMTAB_H_
// public interface
#include "symtab.h"
#endif

// See symtab.h for descriptionm

#ifndef _SYMTAB_PRIV_H_
// private section
#include "symtab_priv.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// #define TEST_MAIN 1   // Generate a main() to test the library - set in Makefile

#ifndef TRUE
#define TRUE (0==0)
#define FALSE (0!=0)
#endif

static int symtab_debug = FALSE;

int *switchmap = NULL;
int switchmap_arraysize = 0; // system data holding allocated space
int switchmap_next_free = 0;  // index of next free slot for user data
#define switchmap(x) FLEX(switchmap, x, __LINE__)

int new_switch(char *name, int declp, int low, int high) {
  int i, default_sw = switchmap_next_free;
  // slot for low index, high index, default used?
  switchmap_next_free = 3 + switchmap_next_free + high - low + 1; // slots for low:high inclusive used?

  makespace(switchmap, switchmap_next_free+1);
  
  for (i = default_sw; i <= switchmap_next_free; i++) { // initialise all flags
    switchmap(i) = 0; // 0: label not seen.  != 0: label used.
  }
  switchmap(default_sw) = low;
  switchmap(default_sw+1) = high;
  switchmap(default_sw+2) = 0; // default used?
  add_entry("Switches", name, default_sw);
  //fprintf(stderr, "Declared %%switch %s(%d:%d)\n", name, low, high);
  return default_sw;
}

void mark_switch_case_used(char *names, int val) {
  int switch_info = lookup("Switches", names);
  int low  = switchmap(switch_info+0);
  int high = switchmap(switch_info+1);
  // +2 is the default flag
  // +3 is the first element of low:high
  if (val < low || val > high) {
    fprintf(stderr, "Switch %s(%d): out of range (%d:%d)\n", names, val, low, high);
    // This should be moved to compile()! (as a fault())
  } else {
    switchmap(switch_info+3+val-low) = 1;
    //fprintf(stderr, "Used switch %s(%d): (index %d)\n", names, val, val-low);
  }
}

void mark_switch_default_used(char *names) {
  int switch_info = lookup("Switches", names);
  switchmap(switch_info+2) = 1; // default has been used.
  //fprintf(stderr, "Used switch %s(*):\n", names);
}


scope_list *top_level_scope = NULL;
scope_list *base_level_scope = NULL;

static void *Malloc(int n) {
  void *p = malloc(n);
  if (symtab_debug) fprintf(stderr, "Malloc: %p\n", p);
  return p;
}

static char *Strdup(char *s) {
  char *p = strdup(s);
  if (symtab_debug) fprintf(stderr, "Strdup: %p (%s)\n", p, s);
  return p;
}

static void Free(void *p) {
  if (symtab_debug) fprintf(stderr, "Free: %p\n", p);
}

// Using 'match' rather than strcmp in the lookup procedures allows us to look up a
// variable in multiple symbol tables.  We also sneak in a wildcard here though it
// is unlikely to ever be very useful with the current interface.

static int match(char *given_name, char *symbol_name) {
  if (given_name == symbol_name) return TRUE; // fast match if using strings found in a string table.
  if (*given_name == '*' && given_name[1] == '\0') return TRUE; // wildcard match. Only "*" supported.
  return strcmp(given_name, symbol_name) == 0;
}

static void destroy_table(generic_table_entry *entry) {
  generic_table_entry *next_entry;
  while (entry != NULL) {
    next_entry = entry->next_entry;
    // destroy_data(entry->entry_data); // This would require a callback... let's leave it to the user for now.

    // If we want to undef a macro that was wrapping an array, we would do so here...
    // and also restore any macros of the same name for an enclosing scope level that
    // had been hidden by this one.  However doing so here is *not* the way to do it!
    // If we go that route, we need to add a callback on deletion.  Which could do
    // double service with destroying the associated Sym data, if need be.

    if (symtab_debug) fprintf(stderr, "Removing %s from scope %d\n", entry->entry_name, current_scope+1);
    Free(entry->entry_name); // strdup & free needed due to use of strpool strings which are volatile.

    Free(entry);
    entry = next_entry;
  }
}

static scope_list *new_scope(scope_list *previous_scopes) {
  scope_list *scope = Malloc(sizeof(scope_list));
  scope->level = ++current_scope;
  scope->table = NULL;
  scope->next_scope = previous_scopes;
  if (symtab_debug) fprintf(stderr, "Create %s scope %d\n", (previous_scopes == NULL ? "top-level" : "nested"), current_scope);
  return scope;
}

static void destroy_tables(table_list *table) {
  table_list *next_table;
  while (table != NULL) {
    next_table = table->next_table;
    destroy_table(table->table_data);
    Free(table);
    table = next_table;
  }
}

static void end_scope(scope_list **parent) {
#define scope (*parent)
  scope_list *next_scope;
  if (parent == NULL || scope == NULL) {
    fprintf(stderr, "Warning! Asked to destroy NULL scope\n");
    return;
  }
  if (scope->level != current_scope) {
    fprintf(stderr, "Warning! Asked to destroy scope level %d, but current scope is %d\n", scope->level, current_scope);
  }
  if (scope != top_level_scope) {
    fprintf(stderr, "Warning! Scope being ended (level %d) is not the top-level scope (level %d)!\n", scope->level, top_level_scope->level);
    return;
  }
  if (symtab_debug) fprintf(stderr, "Destroying scope level %d\n", current_scope);
  next_scope = scope->next_scope;
  current_scope -= 1;
  destroy_tables(scope->table); scope->table = NULL;
  Free(scope);
  scope = next_scope;
#undef scope
}

static void add_new_table(scope_list *current_scope_list, char *table_name) {
  table_list *table = Malloc(sizeof(table_list));
  table->table_name = table_name;
  table->table_data = NULL;
  table->next_table = current_scope_list->table;
  current_scope_list->table = table;
}

static int add_entry_internal(char *table_name, char *entry_name, int entry_data_idx, int top) {
  // If table_name does not exist at this scope, create it.
  scope_list *s = top_level_scope;
  if (symtab_debug && top) fprintf(stderr, "Add %s/%s to Scope: %d\n", table_name, entry_name, s->level);
  if (s->table != NULL) {
    table_list *t = s->table;
    while (t != NULL) {
      if (symtab_debug && top) fprintf(stderr, "  Add to Table: %s\n", t->table_name);
      if (match(table_name, t->table_name)) {
        // Table exists!
        // add entry to the table in current scope

        // Check for presence:
        generic_table_entry *d = t->table_data;
        if (top) {
          while (d != NULL) {
            if (symtab_debug && top) fprintf(stderr, "    Add to Entry: %s\n", d->entry_name);
            if (match(entry_name, d->entry_name)) {
              // User can warn or error if already present...
              return FALSE;
            }
            d = d->next_entry;
          }
        }

        // Now add:
        d = Malloc(sizeof(generic_table_entry));
        d->entry_name = Strdup(entry_name); // unfortunately the stringpool strings being passed to us are transient due to flex array being reallocated
                                            // would be OK if we stored integer indexes instead of actual strings.
        d->entry_data = entry_data_idx;
        d->next_entry = t->table_data;
        t->table_data = d;
        return TRUE;
      }
      t = t->next_table;
    }
  }
  // Table does not exist!  Create it!
  add_new_table(top_level_scope, table_name);
  return add_entry_internal(table_name, entry_name, entry_data_idx, FALSE); // Safe to unconditionally add now...
}

int add_entry(char *table_name, char *entry_name, int entry_data_idx) {
  return add_entry_internal(table_name, entry_name, entry_data_idx, TRUE);
}

// exported, but not included in header file.
int lookup_with_scope(char *table_name, char *entry_name, int *scope_level) {
  scope_list *s = top_level_scope;
  while (s != NULL) {
    if (symtab_debug) fprintf(stderr, "Lookup %s/%s in Scope: %d\n", table_name, entry_name, s->level);
    if (s->table != NULL) {
      table_list *t = s->table;
      if (symtab_debug) fprintf(stderr, "  Lookup %s in Table: %s\n", table_name, t->table_name);
      while (t != NULL) {
        if (match(table_name, t->table_name)) {
          generic_table_entry *d = t->table_data;
          while (d != NULL) {
            if (symtab_debug) fprintf(stderr, "    Lookup %s in Entry: %s\n", entry_name, d->entry_name);
            if (match(entry_name, d->entry_name)) {
              if (symtab_debug) fprintf(stderr, "    Match %s in Entry: %s - scope level = %d\n", entry_name, d->entry_name, s->level);
              if (scope_level != NULL) *scope_level = s->level;
              return d->entry_data;
            }
            d = d->next_entry;
          }
        }
        t = t->next_table;
      }
    }
    s = s->next_scope;
  }
  return -1;
}

int lookup(char *table_name, char *entry_name) {
  return lookup_with_scope(table_name, entry_name, NULL);
}

int scope_level(char *table_name, char *entry_name) {
  int result;
  int scope_level;
  result = lookup_with_scope(table_name, entry_name, &scope_level);
  return scope_level; // which may be -1
}

void push_scope_level(void) {
  if (top_level_scope == NULL) {
    top_level_scope = base_level_scope = new_scope(NULL); // Move to scope level 0
    return;
  }
  top_level_scope = new_scope(top_level_scope); // Move to scope level 1
}

/*
typedef struct generic_table_entry {
  char *entry_name;
  int   entry_data;
  generic_table_entry *next_entry;
} generic_table_entry;

typedef struct table_list {
  char *table_name;
  generic_table_entry *table_data;
  table_list *next_table;
} table_list;

typedef struct scope_list {
  int level;
  table_list *table;
  scope_list *next_scope;
} scope_list;
 */

void (*callback)(char *tablename, char *entryname, int entrydata, void *userdata) = NULL;
void *callbackparam = NULL;

void install_callback(void (*user_callback)(char *tablename, char *entryname, int entrydata, void *userdata), void *userdata) {
  callback = user_callback; callbackparam = userdata;
}

void remove_callback(void) {
  callback = NULL; callbackparam = NULL;
}

void pop_scope_level(void) {
  if (callback != NULL) {
    table_list *table;
    table = top_level_scope->table;
    while (table != NULL) {
      generic_table_entry *entry_list;
      entry_list = table->table_data;
      while (entry_list != NULL) {
        callback(table->table_name, entry_list->entry_name, entry_list->entry_data, callbackparam);
        entry_list = entry_list->next_entry;
      }
      table = table->next_table;
    }
  }
  end_scope(&top_level_scope);
  if (top_level_scope == NULL) base_level_scope = NULL;
}

#ifdef TEST_MAIN
int main(int argc, char **argv) {
  double pi = 3.141592653589793;
  double pi2 = 3.14;
  double *found;

  if ((argc == 2) && (strcmp(argv[1], "-d") == 0)) symtab_debug = TRUE;

  push_scope_level(); // Move to scope level 0
  fprintf(stderr, "push scope\n");
  
  // declare any over-ridable extrinsics here...
  push_scope_level(); // Move to scope level 1 - ready for externals, top-level statics, etc.
  fprintf(stderr, "push scope\n");

  // enter a '{' group
  push_scope_level(); // Move to next scope level - inside a '{ ... }' block
  fprintf(stderr, "push scope\n");

  if (!add_entry("floats", "pi", &pi)) fprintf(stderr, "Error: %s already exists in %s in scope %d!\n", "pi", "floats", current_scope);
  else fprintf(stderr, "Added floats/pi at scope %d\n", current_scope);

  // enter a '{' group
  push_scope_level(); // Move to next scope level - inside a '{ ... }' block
  fprintf(stderr, "push scope\n");

  if (!add_entry("floats", "pi", &pi2)) fprintf(stderr, "Error: %s already exists in %s in scope %d!\n", "pi", "floats", current_scope);
  else fprintf(stderr, "Added floats/pi at scope %d\n", current_scope);

  if (!add_entry("floats", "pi", &pi2)) fprintf(stderr, "Error: %s already exists in %s in scope %d!\n", "pi", "floats", current_scope);
  else fprintf(stderr, "Added floats/pi at scope %d\n", current_scope);

  found = lookup("floats", "pi");
  if (found) fprintf(stdout, "      pi in scope %d is %f\n", current_scope, *found);
  else fprintf(stderr, "      floats/pi not found in scope %d\n", current_scope);
  // return from '}' group
  pop_scope_level();
  fprintf(stderr, "pop scope\n");

  found = lookup("floats", "pi");
  if (found) fprintf(stdout, "      pi in scope %d is %f\n", current_scope, *found);
  else fprintf(stderr, "      floats/pi not found in scope %d\n", current_scope);
  // return from '}' group
  pop_scope_level();
  fprintf(stderr, "pop scope\n");

  found = lookup("floats", "pi");
  if (found) fprintf(stdout, "      pi in scope %d is %f\n", current_scope, *found);
  else fprintf(stderr, "      floats/pi not found in scope %d\n", current_scope);
  pop_scope_level();
  fprintf(stderr, "pop scope\n");

  while (current_scope >= 0) {
    pop_scope_level(); // Clean up...
    fprintf(stderr, "pop scope\n");
  }
  
  return 0;
}
#endif
